(in-package :cells)

(defun dump-db (db)
  (let ((stats (count-db db)))
    (print-db-stats stats)))

(defun count-db (db)
  (let ((dirs (get db 'directories)))
    (if dirs
        (loop for (nil . subdb) in dirs
              collecting (count-db subdb) into dir-stats
              finally (return (list db
                                (reduce (lambda (s1 s2)
                                          (mapcar '+ s1 s2))
                                  (mapcar 'second dir-stats))
                                (sort dir-stats '> :key 'caadr))))
      (count-dir db))))

(defun db-assoc (db keys)
  (loop for key in keys collecting (cons key (get db key 0))))

(defun count-dir (db)
  (loop for (fname . fdb) in (get db 'files)
      for lines = (get fdb 'line-count)
      for chars = (get fdb 'char-count)
      do (format nil *dump*
           fname
           chars lines)
      summing lines into dir-lines
      summing chars into dir-chars
      do (format t "~&~a ~{~a~} ~{~a~} ~{~a~}"
           fname (db-assoc fdb *c-making*)
           (db-assoc fdb *c-defining*)
           (db-assoc fdb *c-defining*))
      finally (return (list db (list (floor dir-chars 1000) dir-lines)))))

(defun print-db-stats (db-stats &optional (depth 0))
  (print (list depth db-stats)))

#+clinisys
(#:|cs2001| (1285 120228)
  ((#:|application| (975 91332)
    ((#:|graphics| (358 34448)
      ((#:|MDGraphics| (198 18114))
       (#:|MGTextEdit| (98 11400))
       (#:|OSGraphics| (62 4934))))
     (#:|configuration| (225 21966)
      ((#:|Forms| (168 16020))
       (#:|TrialBuilder| (40 3756))
       (#:|FormsCompiler| (17 2190))))
     (#:|browser| (165 14842))
     (#:|scanning| (92 7972)
      ((#:|CSRecognize| (54 4546))
       (#:|Axtel| (21 1638))
       (#:|FormFix| (9 918))
       (#:|CSScan| (8 870))))
     (#:|database| (79 7150)
      ((#:|DBModel| (42 3800))
       (#:|DBSynch| (37 3350))))
     (#:|utilities| (56 4954))))
   (#:|dataflow| (138 12970)
    ((#:|Model| (82 7864))
     (#:|PSModel| (56 5106))))
   (#:|provent| (93 8000))
   (#:|CliniSys| (79 7926))))

#+not
(defun print-db-stats (db &optional (depth 0))
  (let ((dirs (get db 'directories)))
    (if dirs
        (loop for (nil . subdb) in dirs
              collecting (dump-db subdb (1+ depth)) into dir-stats
              finally (let ((db-stats (reduce (lambda (s1 s2)
                                        (mapcar '+ s1 s2)) dir-stats)))
                        (apply 'format t *dump*
                          (make-string depth :initial-element #\|)
                          db db-stats)
                        (return db-stats)))
      (dump-dir db depth))))

#+yet
(defun count-db (db depth)
  (loop for (fname . fdb) in (get db 'files)
      for lines = (get fdb 'line-count)
      for chars = (get fdb 'char-count)
      do (format nil *dump*
           (make-string depth :initial-element #\|)
           fname
           chars lines)
      summing lines into dir-lines
      summing chars into dir-chars
      finally (format t *dump*
                (make-string depth :initial-element #\|)
                db
                (floor dir-chars 1000) dir-lines)
        (return (list (floor dir-chars 1000) dir-lines))))

